package goaleditor;


/*************************************************************************

  Copyright (c) 2005 Universita degli Studi di Trento All Rights Reserved.

This file contains software and/or modifications that are subject to the
University of Trento Shared Source License Version 1.0 (the 'License')
restricting usage for non-commercial purposes.

If this file contains source code, the licence should be available below
otherwise a copy of the License is available at http://www.dit.unitn.it/ssl

*************************************************************************/

/* Universita degli Studi di Trento ("UNITN") - Shared Source Licence Version 1.0 **

This License governs use of the accompanying Software, and your use of the Software
constitutes acceptance of this license.

You may use this Software for any non-commercial purpose, subject to the restrictions
in this license. Some purposes which can be non-commercial are teaching, academic
research, and personal experimentation. You may also distribute this Software with
books or other teaching materials, or publish the Software on websites, that are
intended to teach the use of the Software.

You may not use or distribute this Software or any derivative works in any form
for commercial purposes under this license. Examples of commercial purposes would be running business
operations, licensing, leasing, or selling the Software, or distributing the Software
for use with commercial products.

You may modify this Software and distribute the modified Software for non-commercial
purposes, however, you may not grant rights to the Software or derivative works that
are broader than those provided by this License. For example, you may not distribute
modifications of the Software under terms that would permit commercial use, or under
terms that purport to require the Software or derivative works to be sublicensed
to others.

You may use any information in intangible form that you remember after accessing
the Software. However, this right does not grant you a license to any of
UNITN's copyrights or patents for anything you might create using such information.

In return, we simply require that you agree:

- Not to remove any copyright or other notices from the Software.


- That if you distribute the Software in source or object form,
  you will include a verbatim copy of this license.


- That if you distribute derivative works of the Software in source code form
  you do so only under a license that includes all of the provisions of this
  License, and if you distribute derivative works of the Software solely in object
  form you do so only under a license that complies with this License.


- That if you have modified the Software or created derivative works, and distribute
  such modifications or derivative works, you will cause the modified files to carry
  prominent notices so that recipients know that they are not receiving the original
  Software. Such notices must state: (i) that you have changed the Software; and
  (ii) the date of any changes.


- THAT THE SOFTWARE COMES "AS IS", WITH NO WARRANTIES. THIS MEANS NO EXPRESS, IMPLIED
  OR STATUTORY WARRANTY, INCLUDING WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY
  OR FITNESS FOR A PARTICULAR PURPOSE OR ANY WARRANTY OF TITLE OR NON-INFRINGEMENT.
  ALSO, YOU MUST PASS THIS DISCLAIMER ON WHENEVER YOU DISTRIBUTE THE SOFTWARE OR
  DERIVATIVE WORKS.


- THAT UNITN WILL NOT BE LIABLE FOR ANY DAMAGES RELATED TO THE SOFTWARE OR THIS
  LICENSE, INCLUDING DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL OR INCIDENTAL DAMAGES,
  TO THE MAXIMUM EXTENT THE LAW PERMITS, NO MATTER WHAT LEGAL THEORY IT IS BASED ON.
  ALSO, YOU MUST PASS THIS LIMITATION OF LIABILITY ON WHENEVER YOU DISTRIBUTE THE
  SOFTWARE OR DERIVATIVE WORKS.


- That if you sue anyone over patents that you think may apply to the Software or
  anyone's use of the Software, your license to the Software ends automatically.


- That your rights under the License end automatically if you breach it in any way.


UNITN may, at its sole discretion, choose to license any modifications made by or
on behalf of UNITN under this License, or on different terms from those
contained in this license or may choose not to license them at all.

UNITN reserves all rights not expressly granted to you in this license.

********************************************************************************/

/**
 * @author Alberto Siena
 */

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.io.*;
import java.text.AttributedString;
import java.text.AttributedCharacterIterator;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;


interface GENodeShape
{
  void draw( Graphics2D g, int rx, int ry, int rw, int rh );
  //Shape getShape( Rectangle r );
  Shape getShape( int rx, int ry, int rw, int rh );
  String getName();
}



public class GENode extends GEObject implements GEMoveable
{
  static class DefaultDataModel extends GEDataModel
  {
    protected ArrayList m_relations = new ArrayList();

    public void addRelation( GEDataModel dm )
    {
      if( !(dm instanceof GERelation.DataModel ) ) return;

      Integer id = new Integer( (int)dm.getGuid() );

      if( m_relations.contains( id ) ) return;

      m_relations.add( new Integer( (int)dm.getGuid() ) );
    }
    public void removeRelation( GEDataModel dm )
    {
      m_relations.remove( new Integer( dm.getGuid() ) );
    }
    public void removeRelation( int r )
    {
      m_relations.remove( new Integer( r ) );
    }
    public ArrayList enumRelations()
    {
      return m_relations;
    }
    public void addFilter( ModelFilter mf, boolean b )
    {
      if( b )
      {
        m_modelFilters.remove( mf );
      }
      else
      {
        if( !m_modelFilters.contains( mf ) )
        {
          m_modelFilters.add( mf );
        }
      }
    }
  }


  static class GUI
  {
    static class Label extends JLabel
    {
      public Label()
      {
        init( null );
      }
      public Label( int id )
      {
        init( (GENode.DefaultDataModel)GEApp.getDocument().findGoal( id ) );
      }
      public Label( GENode node )
      {
        init( (GENode.DefaultDataModel)node.getData() );
      }
      protected void init( GENode.DefaultDataModel node )
      {
        setOpaque( true );
        setBackground( Color.WHITE );
        setBorder( BorderFactory.createLoweredBevelBorder() );
        if( node != null ) setText( node.getCaption() );
        else setText( "-" );
      }
    }
  }


  //**************************************************************************
  //**************** Graph Data **********************************************
  //**************************************************************************

  //private Color m_color = GEObject.defFillColor;
  protected Font m_font = new Font( "Arial", Font.PLAIN, 14 );
  //boolean m_colorAuto = true;

  protected ArrayList m_connectors = new ArrayList();
  protected GETrackList m_handles = new GETrackList();
  protected boolean m_lockResizing = true;

  protected AttributedString m_attributedString;
  private Rectangle m_rect = new Rectangle();

  public static final PropertiesPanel propPage = new PropertiesPanel();


  //**************************************************************************
  //*********** Static data **************************************************
  //**************************************************************************

  static final String[] opLabel = { "AND", "OR" };
  public static boolean ShowSolverValues = false;


  public static final int objetTypeNODE = 1;

  public static Color COLOR_DARK_GREEN = new Color( 0, 88, 0 );
  public static Color COLOR_LINKED_NODE = new Color( 0xff, 0xdd, 0xdd );


  //**************************************************************************
  //************* Costruttori ************************************************
  //**************************************************************************

  public GENode()
  {
    super();
    init();
  }
  public GENode( int t )
  {
    super();
    init();
  }
  public GENode( int ix1, int iy1, int ix2, int iy2 )
  {
    super();
    this.setBounds( ix1, iy1, ix2 - ix1, iy2 - iy1);
  }
  public GENode( int t, int ix1, int iy1, int ix2, int iy2 )
  {
    super();
    this.setBounds( ix1, iy1, ix2 - ix1, iy2 - iy1);
  }

  public GEObject init()
  {
    return super.init();
  }

  public GEDataModel createData()
  {
    return new GEDataModel();
  }

  //**************************************************************************
  //*************** Property get / set ***************************************
  //**************************************************************************

   public boolean is( Class cls )
   {
     return false;
   }

  public GEDataModel getData()
  {
    return null;
  }
  public void setData( GEDataModel data )
  {
  }

  public GERelation.Connector getConnector( int n )
  {
    return (GERelation.Connector)m_connectors.get( n );
  }

  public boolean isForeColorAuto()
  {
    return m_foreColor == GEObject.defForeColor;
  }
  public boolean isFillColorAuto()
  {
    return m_fillColor == GEObject.defFillColor;
  }

  public ArrayList getConnectors()
  {
    return m_connectors;
  }

  public ArrayList enumOutgoingConnectors()
  {
    ArrayList list = new ArrayList();

    for( int i = 0; i < m_connectors.size(); i++ )
    {
      GERelation.Connector ctn = getConnector( i );
      if( ctn.startg == this ) list.add( ctn );
    }

    return list;
  }
  public ArrayList enumIncomingConnectors()
  {
    ArrayList list = new ArrayList();

    for( int i = 0; i < m_connectors.size(); i++ )
    {
      GERelation.Connector ctn = getConnector( i );
      if( ctn.endg == this ) list.add( ctn );
    }

    return list;
  }

  public Color getFillColor()
  {
    /*if( isFillColorAuto() )
    {
      return defFillColor.getColor();
    }*/

    return m_fillColor.getColor();
  }
  public Color getForeColor()
  {
    //if( m_selected ) return Color.RED;

    return m_foreColor.getColor();
  }

  public void lockResizing( boolean b )
  {
    m_lockResizing = b;
  }
  public boolean lockResizing()
  {
    return m_lockResizing;
  }

  public XmlNode createDescriptor()
  {
    return new XmlNode( "NODE" );
  }
  public XmlNode getDescriptor()
  {
    XmlNode d = super.getDescriptor();

    {
      XmlNode bd = d.add( new XmlNode( "BOUNDS" ) );
      bd.add( "X", getX() );
      bd.add( "Y", getY() );
      bd.add( "W", getWidth() );
      bd.add( "H", getHeight() );
    }
    {
      XmlNode ndColor = d.add( new XmlNode( "COLOR" ) );
      ndColor.add( "FOREGROUND", getForeColor().getRGB() );
      ndColor.add( "BACKGROUND", getFillColor().getRGB() );
    }
    d.add( "VISIBLE", isVisible() );
    try
    {
      d.add( "DATA", (int)getData().getGuid() );
    }
    catch( NullPointerException ex )
    {
      //System.out.println( "Null pointer exception in actor " + getId() );
    }
    d.add( "ID", getId() );
    d.add( "LOCKRESIZE", lockResizing() );
    //d.add( "DESCRIPTION", getDescription() );

    return d;
  }
  public void setDescriptor( XmlNode nd )
  {
    setId( nd.item( "ID" ).getInt() );

    {
      GEDataModel g = GEApp.getDocument().findGoalByUid( nd.item( "DATA" ).getInt( 0 ) );
      if( g != null )
      {
        setData( g );
      }
      //else { setCaption( " " ); }
    }

    {
      XmlNode xBounds = nd.item( "BOUNDS" );
      if( xBounds.exists() )
      {
        setBounds( xBounds.item( "X" ).getInt( 0 ), xBounds.item( "Y" ).getInt( 0 ),
                   xBounds.item( "W" ).getInt( 100 ), xBounds.item( "H" ).getInt( 60 ) );
      }
    }

    lockResizing( nd.item( "LOCKRESIZE" ).getBoolean( false ) );
    setVisible( nd.item( "VISIBLE" ).getBoolean( true ) );
  }

  //**************************************************************************
  //*************** Eventi ***************************************************
  //**************************************************************************

   public void evtGroupDragged( Point delta )
   {
     move( getX() + delta.x, getY() + delta.y );

     for( int i = 0; i < m_connectors.size(); i++ )
     {
       ((GERelation.Connector)m_connectors.get(i)).evtNodeDragged( this );
     }
   }

    public void evtHandleMoved( GEHandle h, MouseEvent e )
    {
      switch ( h.mnemonic )
      {
        case 0:
          this.setBounds( getX() + h.deltaX(), getY() + h.deltaY(),
                          getWidth() - h.deltaX(), getHeight() - h.deltaY() );
          break;
        case 1:
          this.setBounds( getX(), getY() + h.deltaY(), getWidth(),
                          getHeight() - h.deltaY() );
          break;
        case 2:
          this.setBounds( getX(), getY() + h.deltaY(), getWidth() + h.deltaX(),
                          getHeight() - h.deltaY() );
          break;
        case 3:
          this.setBounds( getX() + h.deltaX(), getY(), getWidth() - h.deltaX(),
                          getHeight() );
          break;
        case 4:
          this.setBounds( getX(), getY(), getWidth() + h.deltaX(), getHeight() );
          break;
        case 5:
          this.setBounds( getX() + h.deltaX(), getY(), getWidth() - h.deltaX(),
                          getHeight() + h.deltaY() );
          break;
        case 6:
          this.setBounds( getX(), getY(), getWidth(), getHeight() + h.deltaY() );
          break;
        case 7:
          this.setBounds( getX(), getY(), getWidth() + h.deltaX(),
                          getHeight() + h.deltaY() );
          break;
      }
      if( lockResizing() )
      {
        if( getWidth() < getMinimalWidth() )
        {
          setSize( getMinimalWidth(), getHeight() );
          e.translatePoint( -h.deltaX(), 0 );
        }
        if( getHeight() < getMinimalHeight() )
        {
          setSize( getWidth(), getMinimalHeight() );
          e.translatePoint( 0, -h.deltaY() );
        }
      }
      recalc();
    }

    public void evtMouseDragged( MouseEvent e )
    {
      super.evtMouseDragged( e );
      recalc();
    }

    public void evtMouseEntered(MouseEvent e)
    {
      showTooltip();
    }

    protected void showTooltip()
    {
      if ( getDescription().length() < 1 )
        return;
      GEPageTooltip m_tooltip = getView().getTooltip();
      m_tooltip.setText( getDescription() );
      m_tooltip.setBounds(
          getX() + getWidth() / 2, getY() + getHeight() - 5,
          150, 50 );

      m_tooltip.show();
    }

    public void evtMouseExited(MouseEvent e)
    {
      if( getView() == null )
      {
        //System.out.println();
      }
      GEPageTooltip m_tooltip = getView().getTooltip();

      if( m_tooltip != null ) m_tooltip.hide();
    }

    public void evtMousePressed( MouseEvent e )
    {
      super.evtMousePressed( e );

      if( getDescription().length() < 1 ) return;

      getView().getTooltip().hide();
    }

    public void evtMouseReleased( MouseEvent e )
    {
      PropertiesPanel pnl = ( PropertiesPanel ) GEApp.getFrame().getSidePanel();
      pnl.txtCaption.setSelectionStart( 0 );
      pnl.txtCaption.setSelectionEnd( pnl.txtCaption.getText().length() );
      pnl.txtCaption.requestFocus();

      super.evtMouseReleased( e );

      showTooltip();
    }

    public void evtConnectorRemoved( GERelation.Connector r )
    {
      ((DefaultDataModel)getData()).removeRelation( r.getRelation() );
      m_connectors.remove( r );
    }
    public void evtConnectorAdded( GERelation.Connector r )
    {
      ((GENode.DefaultDataModel)getData()).addRelation( r.getData() );
    }

    public void evtRelationTypeChanged( GERelation.Connector r )
    {
    }


  //**************************************************************************
  //**************************************************************************
  //*************** Metodi ***************************************************
  //**************************************************************************

  //**************************************************************************
  //*************** Utilita' *************************************************
  //**************************************************************************

  int calcAngle( int vx1, int vy1, int vx2, int vy2 )
  {
    double lengthX = vx2 - vx1;
    double lengthY = vy2 - vy1;

    // Calulate the length of the line
    double lineLength = Math.sqrt (lengthX * lengthX + lengthY * lengthY);

    int startAngle = (int)Math.toDegrees(Math.asin(lengthY/lineLength));

    if (lengthX > 0) startAngle = (180 - startAngle);

    return startAngle;
  }

  protected void render( Graphics2D g, Rectangle rect )
  {
    g.setColor( Color.BLACK );
    g.drawRect( 0, 0, getWidth(), getHeight() );
    g.drawLine( 0, 0, getWidth(), getHeight() );
    g.drawLine( getWidth(), 0, 0, getHeight() );
  }
  public void print( Graphics2D g, GEPrintInfo pi )
  {
    render( g, new Rectangle( getX() - pi.area.x, getY() - pi.area.y, (int)(getWidth() * pi.zoom), (int)(getHeight() * pi.zoom) ) );
  }
  public void paint(Graphics g )
  {
    render( (Graphics2D)g, new Rectangle( 0, 0, getWidth(), getHeight() ) );
  }

  void drawCaption( Graphics2D g, AttributedString s, Rectangle r )
  {
    AttributedCharacterIterator it = s.getIterator();
    int paragraphStart = it.getBeginIndex();
    int paragraphEnd = it.getEndIndex();
    LineBreakMeasurer lineMeasurer = new LineBreakMeasurer( it,
        g.getFontRenderContext() );
                          //new FontRenderContext(null, false, false));

    Dimension size = r.getSize();
    float formatWidth = (float) size.width - 10;

    float drawPosY = 0;

    lineMeasurer.setPosition(paragraphStart);
    int lines = 0;
    int sz = 0;
    while (lineMeasurer.getPosition() < paragraphEnd)
    {
      lines++;
      TextLayout layout = lineMeasurer.nextLayout(formatWidth);
      sz += layout.getAscent();
    }

    //System.out.println( sz );
    lineMeasurer.setPosition(0);
      while (lineMeasurer.getPosition() < paragraphEnd)
      {

            TextLayout layout = lineMeasurer.nextLayout(formatWidth);
            drawPosY += layout.getAscent();

            float drawPosX;
            drawPosX = formatWidth - layout.getAdvance();

            layout.draw( g,
                         (size.width /2) - (layout.getAdvance() /2),
                         (size.height /2) + drawPosY - (sz /2)- layout.getDescent() );

            drawPosY += layout.getDescent() + layout.getLeading();
        }
  }

  public void setVisible( boolean b, boolean rels )
  {
    setVisible( b );

    if( rels )
    {
      for( int r = 0; r < m_connectors.size(); r++ )
      {
        GERelation.Connector rel = (GERelation.Connector)m_connectors.get( r );
        boolean gv = true;
        if( rel.endg != this )
        {
          gv = rel.endg.isVisible();
        }
        if( rel.startg != this )
        {
          gv = rel.startg.isVisible();
        }
        if( b && gv )
         {
           rel.recalcLine();
           rel.setVisible( true );
         }
        if( !b ) rel.setVisible( false );
      }
    }
  }


  void recalc()
  {
    GERelation.Connector r;
    for( int i = 0; i < m_connectors.size(); i++ )
    {
      getView().ensureBounds( getX() + getWidth(), getY() + getHeight() );
      r = (GERelation.Connector)m_connectors.get(i);
      r.evtNodeDragged( this );
    }
    m_rect.width = getWidth();
    m_rect.height = getHeight();
  }

  public boolean canHaveFocus()
  {
    return true;
  }

  public void addConnector( GERelation.Connector r )
  {
    m_connectors.add( r );
    //m_isTop = !hasOutgoingRelations();
    //m_isLeaf = !hasIncomingRelations();
  }
  public void addConnector( GERelation.Connector r, boolean comp )
  {
    m_connectors.add( r );
    //m_isTop = !hasOutgoingRelations();
    //m_isLeaf = !hasIncomingRelations();
  }

  public boolean hasIncomingConnectors()
  {
    for( int j = 0; j < m_connectors.size(); j++ )
    {
      GERelation.Connector r = (GERelation.Connector)m_connectors.get( j );

      if ( r.endg != null )
      {
        if ( getData().getGuid() == r.endg.getData().getGuid() )
        {
          return true;
        }
      }
    }

    return false;
  }
  public boolean hasOutgoingConnectors()
  {
    for ( int j = 0; j < m_connectors.size(); j++ )
    {
      GERelation.Connector r = ( GERelation.Connector ) m_connectors.get( j );

      if ( getData().getGuid() == r.startg.getData().getGuid() )
      {
        return true;
      }
    }

    return false;
  }
  public boolean hasVisibleConnectors()
  {
    for ( int j = 0; j < m_connectors.size(); j++ )
    {
      GERelation.Connector r = ( GERelation.Connector ) m_connectors.get( j );

      if ( r.isVisible() )
      {
        return true;
      }
    }

    return false;
  }

  public Shape getShape()
  {
    return new Rectangle( getX(), getY(), getWidth(), getHeight() );
  }
  public void removeSelf( boolean remove_data )
  {
    GERelation.Connector r;

    if( remove_data ) GEApp.getDocument().remove( getData() );

    if( m_handles.list.size() > 0 ) m_handles.clear();

    int bound = m_connectors.size();
    for( int i = 0; i < bound; i++ )
    {
      ((GERelation.Connector)m_connectors.get(0)).evtNodeRemoved( this );
    }
    super.removeSelf();
  }
  public void removeSelf()
  {
    removeSelf( false );
  }
    public void setCaption( String s )
    {
      if( getData() != null )
      {
        getData().setCaption( s );
        m_attributedString = new AttributedString( s.length() > 0 ? s : " " );
        m_attributedString.addAttribute( TextAttribute.FAMILY, "Helvetica" );
      }
    }
    public String getCaption()
    {
      try
      {
        return getData().getCaption();
      }
      catch( Exception ext )
      {
        ext.printStackTrace();
        return "";
      }
    }
    public String getDescription()
    {
      try
      {
        return getData().getDescription();
      }
      catch( Exception ext )
      {
        ext.printStackTrace();
        return "";
      }
    }
    public void setDescription( String str )
    {
      if( getData() != null ) getData().setDescription( str );
    }

    void collapse()
    {
    }
    boolean isCollapsed()
    {
      return false;
    }
    void expand()
    {
    }
    public int getMinimalWidth()
    {
      return 0;
    }
    public int getMinimalHeight()
    {
      return 0;
    }

    public GEObject.PropertyPanel getPropertiesPanel()
    {
      return new PropertiesPanel( this ).getComponent();
    }

    void evtSelected()
    {
      super.evtSelected();

      showHandles();
    }

    void evtUnselected()
    {
      super.evtUnselected();

      hideHandles();
    }

    void showHandles()
    {
      m_handles.setTarget( this );
      m_handles.setParent( getView() );

      m_handles.addHandle( 0, getX(),
          getY() ).setCursor( Cursor.getPredefinedCursor( Cursor.NW_RESIZE_CURSOR ) );
      m_handles.addHandle( 1, getX() + getWidth() / 2,
          getY() ).setCursor( Cursor.getPredefinedCursor( Cursor.N_RESIZE_CURSOR ) );
      m_handles.addHandle( 2, getX() + getWidth(),
          getY() ).setCursor( Cursor.getPredefinedCursor( Cursor.NE_RESIZE_CURSOR ) );

      m_handles.addHandle( 3, getX(),
          getY() +
          getHeight() /
          2 ).setCursor( Cursor.getPredefinedCursor( Cursor.E_RESIZE_CURSOR ) );
      m_handles.addHandle( 4, getX() + getWidth(),
          getY() +
          getHeight() /
          2 ).setCursor( Cursor.getPredefinedCursor( Cursor.E_RESIZE_CURSOR ) );

      m_handles.addHandle( 5, getX(),
          getY() +
          getHeight() ).setCursor( Cursor.getPredefinedCursor( Cursor.SW_RESIZE_CURSOR ) );
      m_handles.addHandle( 6, getX() + getWidth() / 2,
          getY() +
          getHeight() ).setCursor( Cursor.getPredefinedCursor( Cursor.S_RESIZE_CURSOR ) );
      m_handles.addHandle( 7, getX() + getWidth(),
          getY() +
          getHeight() ).setCursor( Cursor.getPredefinedCursor( Cursor.SE_RESIZE_CURSOR ) );
    }

    void hideHandles()
    {
      m_handles.clear();
    }

    public void setBounds( int _x, int _y, int _w, int _h )
    {
      if( _w < 30 ) _w = 30;
      if( _h < 30 ) _h = 30;

      _x = snapToGrid( _x );
      _y = snapToGrid( _y );
      _w = snapToGrid( _w );
      _h = snapToGrid( _h );

      this.reshape( _x, _y, _w, _h );

      if( this.isSelected() )
      {
        if( m_handles.list.size() > 0 )
        {
          m_handles.getHendle( 0 ).setPoint( _x, _y );
          m_handles.getHendle( 1 ).setPoint( _x + ( _w / 2 ), _y );
          m_handles.getHendle( 2 ).setPoint( _x + _w, _y );
          m_handles.getHendle( 3 ).setPoint( _x, _y + ( _h / 2 ) );
          m_handles.getHendle( 4 ).setPoint( _x + _w, _y + ( _h ) / 2 );
          m_handles.getHendle( 5 ).setPoint( _x, _y + _h );
          m_handles.getHendle( 6 ).setPoint( _x + ( _w ) / 2, _y + _h );
          m_handles.getHendle( 7 ).setPoint( _x + _w, _y + _h );
        }
      }
    }

    public static class PropertiesPanel extends GEObject.PropertyPanel implements FocusListener, ActionListener, YesNoAutoPanelListener
    {
      GENode node;

      JComboBox cmbFillColor = new JComboBox();
      JComboBox cmbForeColor = new JComboBox();
      JTextArea txtDesc = new JTextArea();
      JTextField txtCaption = new JTextField();
      JCheckBox chkLockResizing = new JCheckBox( "Lock resizing" );

      PropertiesPanel(){}
      PropertiesPanel( GENode nd )
      {
        node = nd;
        //initFor( this );
      }

      public GEObject.PropertyPanel getComponent()
      {
        init( this );

        return this;
      }

      public GEObject getObject()
      {
        return node;
      }

      protected void init( Container c )
      {
        c.setLayout( new BorderLayout() );

        c.add( m_tab, BorderLayout.CENTER );

        addPropertyTabs( m_tab );
      }

      protected void addPropertyTabs( JTabbedPane tab )
      {
        tab.addTab( "", new ImageIcon( GEFrame.class.getResource( "prop.png" ) ), getDefaultPanel(), "Visual properties" );
      }

      public Container getDefaultPanel()
      {
        JPanel pnl = new JPanel();

        initFor( pnl );

        return pnl;
      }

      Container initFor( Container c )
      {
        //JLabel label;
        Component northc = null;

        c.setLayout( new SpringLayout() );


          Color paleRed = new Color( 0xffcccc );
          Color paleGreen = new Color( 0xccffcc );
          Color paleBlue = new Color( 0xccccff );
          Color paleYellow = new Color( 0xffffcc );

          addComboBoxItem( cmbForeColor, "[auto]", -1, true );
          addComboBoxItem( cmbForeColor, "Current", node.getForeColor().getRGB(), (node.getPreferredForeColor() != GEObject.defForeColor) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Black", Color.BLACK.getRGB(), (node.getForeColor().getRGB() == Color.BLACK.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Grey", Color.GRAY.getRGB(), (node.getForeColor().getRGB() == Color.GRAY.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Yellow", Color.yellow.getRGB(), (node.getForeColor().getRGB() == Color.yellow.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Red", Color.red.getRGB(), (node.getForeColor().getRGB() == Color.red.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Green", Color.green.getRGB(), (node.getForeColor().getRGB() == Color.green.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Blue", Color.blue.getRGB(), (node.getForeColor().getRGB() == Color.blue.getRGB()) && !node.isForeColorAuto() );
          addComboBoxItem( cmbForeColor, "Dark red", 0x800000, (node.getForeColor().getRGB() == 0x800000) && !node.isForeColorAuto() );

          addComboBoxItem( cmbFillColor, "[auto]", -1, node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Grey", Color.LIGHT_GRAY.getRGB(), (node.getFillColor().getRGB() == Color.LIGHT_GRAY.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "White", Color.WHITE.getRGB(), (node.getFillColor().getRGB() == Color.WHITE.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Yellow", paleYellow.getRGB(), (node.getFillColor().getRGB() == paleYellow.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Red", paleRed.getRGB(), (node.getFillColor().getRGB() == paleRed.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Green", paleGreen.getRGB(), (node.getFillColor().getRGB() == paleGreen.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Blue", paleBlue.getRGB(), (node.getFillColor().getRGB() == paleBlue.getRGB()) && !node.isFillColorAuto() );
          addComboBoxItem( cmbFillColor, "Light grey", 0xeeeeee, (node.getFillColor().getRGB() == new Color( 0xeeeeee ).getRGB()) && !node.isFillColorAuto() );

          chkLockResizing.setSelected( node.lockResizing() );

          northc = addRowComponent( c, "Contour color:", cmbForeColor, northc );
          northc = addRowComponent( c, "Fill color:", cmbFillColor, northc );
          northc = addRowComponent( c, chkLockResizing );

          cmbForeColor.addActionListener( this );
          cmbFillColor.addActionListener( this );
          cmbForeColor.setRenderer( new GECellRenderer() );
          cmbFillColor.setRenderer( new GECellRenderer() );

          txtDesc.setRows( 10 );
          txtDesc.setText( node.getDescription() );
          txtDesc.addFocusListener( this );

          JScrollPane scrDesc = new JScrollPane();
          scrDesc.getViewport().add( txtDesc );
          northc = addRowComponent( c, "Description:", scrDesc, northc );
          txtDesc.setAutoscrolls( true );
          txtDesc.setLineWrap( true );
          txtDesc.setWrapStyleWord( true );
          chkLockResizing.addActionListener( this );
        //}

        return c;
      }

      public void optionChanged( YesNoAutoPanel pnl )
      {
      }

      JComboBox addComboBox( Container cont, String label, JComboBox cmb, Component northc )
      {
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.anchor = GridBagConstraints.WEST;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.weightx = 100;
        northc = addRowComponent( cont, new JLabel( label ), northc );
        northc = addRowComponent( cont, cmb, northc );
        cmb.addActionListener( this );
        return cmb;
      }
      ComboBoxItem addComboBoxItem( JComboBox cmb, String text, int val, boolean sel )
      {
        ComboBoxItem c = new ComboBoxItem( text, val );
        cmb.addItem( c );
        if ( sel == true ) cmb.setSelectedItem( c );

        return c;
      }

      class UpdateAllCaptions implements GEGraphEditor.ForeachObject
      {
        public void doAction( GEObject o )
        {
          if( !(GENode.class.isAssignableFrom( o.getClass() )) ) return;
          ((GENode)o).setCaption( o.getData().getCaption() );
        }
      }

      protected boolean boh = true;
      public void actionPerformed( ActionEvent e )
      {
        if( e.getSource() == txtCaption )
        {
          String text = ( ( JTextField ) e.getSource() ).getText();
          GEDataModel g = GEApp.getDocument().findNodeByCaption( text );

          if( boh == false )
          {
            boh = true;
            node.getView().repaint();
            return;
          }
          boh = false;

          if( GEApp.getConf().shortcuts.datalogUseIds().getBoolean( false ) == false )
          {
            String validchars = "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890_ ";
            for( int i = 0; i < text.length(); i++ )
            {
              if (text.charAt( i ) > 0xff) continue; // Japanease charactor
              if( validchars.indexOf( text.charAt( i ) ) == -1 )
              {
                switch( JOptionPane.showConfirmDialog( GEApp.getFrame(),
                    "Character '" + text.charAt( i ) + "': a node label can only contain\nletters [a-z], numbers, spaces an the underscore character [_] " +
                    "Continue anyway?", "Label format warning", JOptionPane.YES_NO_CANCEL_OPTION ) )
                {
                  case JOptionPane.YES_OPTION:
                    break;
                  case JOptionPane.NO_OPTION:
                  case JOptionPane.CANCEL_OPTION:
                  {
                    return;
                  }
                }
                //GEApp.alert( "Character '" + text.charAt( i ) + "': a node label can only contain\nletters [a-z], numbers, spaces an the underscore character [_] " );
                //return;
              }
            }
          }

          if( (g == null) | (g == node.getData()) )
          {
            node.setCaption( text );
            GEApp.getWorkspace().foreach( GENode.class, new UpdateAllCaptions() );
            this.revalidate();
            //node.repaint();
          }
          else
          {
            boh = false;
            switch ( JOptionPane.showConfirmDialog( GEApp.getFrame(),
                "A node called '" + text + "' already exists.\nContinue anyway?", "ST-Tool",
                JOptionPane.YES_NO_CANCEL_OPTION ) )
            {
              case JOptionPane.YES_OPTION:
              {
                node.setCaption( text );
                GEApp.getWorkspace().foreach( GENode.class, new UpdateAllCaptions() );
                this.revalidate();
              }
              break;

              case JOptionPane.NO_OPTION:
              case JOptionPane.CANCEL_OPTION:
              {
              }
              break;
            }
              /*boh = false;
              switch ( JOptionPane.showConfirmDialog( GEApp.getFrame(),
                  "Node '" + text + "' already exists.\nDo you want to create a link?", "ST-Tool",
                  JOptionPane.YES_NO_CANCEL_OPTION ) )
              {
                case 0:
                {
                  GEDataModel oldg = node.getData();
                  node.setData( g );
                  //System.out.println( oldg.getRefCount() );
                  if( oldg.getRefCount() < 1 )
                  {
                    //GEApp.getDocument().deleteGoal( oldg );
                    GEApp.getDocument().remove( oldg );
                  }
                }
                break;

                case 1:
                case 2:
                {
                  int n = 1;
                  String uname = findFriendlyUName( node.getData(), text );

                  if( uname == null )
                  {
                    while ( GEApp.getDocument().findNodeByCaption( text + " (" + n + ")" ) != null )
                    {
                      n++;
                    }
                    text = text + " (" + n + ")";
                  }
                  else
                  {
                    text = uname;
                  }
                  node.setCaption( text );
                  this.revalidate();
                  node.getView().repaint();
                }
                break;
              }*/
          }
          //node.getParent().repaint();
        }
        /*if( e.getSource() ==  cmbFillColor)
        {
          int val = ((ComboBoxItem)((JComboBox)e.getSource()).getSelectedItem()).toInt();
          if( ((JComboBox)e.getSource()).getSelectedIndex() == 0 )
          //if( val == -1 )
          {
            node.setAutoColor( true );
          }
          else
          {
            node.setAutoColor( false );
            node.setFillColor( new Color( val ) );
          }
        }
        /*if( e.getSource() ==  cmbForeColor)
        {
          int val = ((ComboBoxItem)((JComboBox)e.getSource()).getSelectedItem()).toInt();
          if( ((JComboBox)e.getSource()).getSelectedIndex() == 0 )
          {
            node.setAutoColor( true );
          }
          else
          {
            node.setAutoColor( false );
            node.setFillColor( new Color( val ) );
          }
        }*/
        if( e.getSource() == chkLockResizing )
        {
          node.lockResizing( ((JCheckBox)e.getSource()).isSelected() );
        }
        GEApp.getFrame().getView().evtRefresh();
      }
      protected String findFriendlyUName( GEDataModel g, String text )
      {
        for( int i = 0; i < node.m_connectors.size(); i++ )
        {
          GERelation.Connector ctn = ((GERelation.Connector)node.m_connectors.get( i ));
          if( ctn.getData() instanceof GERelation.DataModel.Composition )
          {
            if( ctn.endg == node )
            {
              String uname = text + " {" + ctn.startg.getCaption() + "}";
              if( GEApp.getDocument().findNodeByCaption( uname ) == null )
              {
                return uname;
              }
              return null;
            }
          }
        }
        return null;
      }
      /*private GEService chooseTargetGoal()
      {
        new GEServiceChooser( GEApp.frame ).chooseGoal( node.getCaption() );
        return null;
      }*/

      public void focusGained( FocusEvent evt )
      {
      }
      public void focusLost( FocusEvent evt )
      {
        if( evt.getSource() == txtDesc )
        {
          node.setDescription( ( ( JTextArea ) evt.getSource() ).getText() );
          //System.out.println( node.getDescription() );
        }
      }
    }

}

class GECellRenderer extends JLabel implements ListCellRenderer {

     public GECellRenderer() {
         setOpaque(true);
     }
     public Component getListCellRendererComponent(
         JList list,
         Object value,
         int index,
         boolean isSelected,
         boolean cellHasFocus)
     {
       ComboBoxItem item = (ComboBoxItem)value;
       setText(value.toString());
       setBackground( new Color(item.toInt()) );
       setForeground( Color.BLACK );
       setBorder( isSelected ? BorderFactory.createLineBorder( Color.BLACK ) : null );
       return this;
     }
 }

class GEServiceChooser extends JDialog implements ActionListener
{
   JButton btnOk = new JButton( "Ok" );
   JButton btnCancel = new JButton( "Cancel" );
   JList m_lstGoals = new JList();
   JCheckBox m_chkLink = new JCheckBox( "Point to an existing goal:" );

   protected boolean m_changes;

   private ArrayList SortArrayList( ArrayList list )
   {
     DuplicateMap map = new DuplicateMap();
     ArrayList ret_list = new ArrayList();

     for( int i = 0; i < list.size(); i++ )
     {
       GEDataModel g = (GEDataModel)list.get( i );
       map.put( g.getCaption(), g );
     }

     Set keys = map.keySet();
     for ( Iterator it = keys.iterator(); it.hasNext(); )
     {
       int n = 0;
       String key = ( String ) it.next();
       GEDataModel g = (GEDataModel)map.get( key );
       while( g != null )
       {
         ret_list.add( g );
         n++;
         g = (GEDataModel)map.get( key, n );
       }
     }
     return ret_list;
   }

   public GEServiceChooser( Frame parent, Class cls )
   {
     Dimension dlgSize = getPreferredSize();
     Rectangle r = parent.getBounds();
     setSize( new Dimension( 300, 200 ) );
     setLocation( r.x + (int)(r.getWidth() /2) - (getWidth() /2), r.y + (int)(r.getHeight() /2) - (getHeight() /2) );
     setModal(true);
     setTitle("STTool :: Select node");

     getContentPane().setBackground(SystemColor.control);
     getContentPane().setLayout( new BorderLayout() );

     JPanel pnlBottom = new JPanel();

     btnOk.setBounds(new Rectangle(194, 198, 115, 44));

     pnlBottom.add( btnOk );
     pnlBottom.add( btnCancel );

     //ArrayList goals = SortArrayList( GEApp.getDocument().getGoals() );
     ArrayList goals = null;
     if( cls == null )
     {
       goals = SortArrayList( GEApp.getDocument().getDataList() );
     }
     else
     {
       goals = SortArrayList( GEApp.getDocument().enum( cls ) );
     }
     ArrayList captions = new ArrayList();
     for( int i = 0; i < goals.size(); i++ )
     {
       GEDataModel g = (GEDataModel)goals.get( i );
       captions.add( g.getCaption() );
     }
     m_lstGoals.setListData( captions.toArray() );
     m_lstGoals.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );

     JScrollPane scr = new JScrollPane();
     scr.getViewport().setView( m_lstGoals );

     getContentPane().add( m_chkLink, BorderLayout.NORTH );
     getContentPane().add( scr, BorderLayout.CENTER );
     getContentPane().add( pnlBottom, BorderLayout.SOUTH );

     btnOk.addActionListener( this );
     btnCancel.addActionListener( this );
     m_chkLink.addActionListener( this );
     enableEvents(AWTEvent.WINDOW_EVENT_MASK);
   }

   public boolean dataChanged()
   {
     return m_changes;
   }

   void ok()
   {
     m_changes = true;
     dispose();
   }
   void cancel()
   {
     m_changes = false;
     dispose();
   }

   public boolean showDialog()
   {
     show();

     if( dataChanged() ) return true;
     else return false;
   }
   public String chooseGoal( GENode node )
   {
     m_lstGoals.setSelectedValue( node.getCaption(), true );
     m_chkLink.setSelected( node.getData().getRefCount() > 1 );

     m_lstGoals.setEnabled( m_chkLink.isSelected() );

     // cancel
     if( showDialog() == false ) return "";

     if( !m_chkLink.isSelected() ) return "[own]";

     return (String)m_lstGoals.getSelectedValue();
   }

   public void actionPerformed( ActionEvent e )
   {
     if( e.getSource() == m_chkLink )
     {
       m_lstGoals.setEnabled( m_chkLink.isSelected() );
     }
     if( e.getSource() == btnOk )
     {
       ok();
     }
     if( e.getSource() == btnCancel )
     {
       cancel();
     }
   }
}

interface GEServiceComboBoxFilter
{
   boolean filterGoal( GEDataModel g );
}

 class GEServiceComboBoxFilter_None implements GEServiceComboBoxFilter
 {
    public boolean filterGoal( GEDataModel g )
    {
      return false;
    }
 }

class GEServiceComboBoxFilter_Caption implements GEServiceComboBoxFilter
{
   String captionStart = "";
   //String captionEnd = "";
   public GEServiceComboBoxFilter_Caption( String s )
   {
     captionStart = s;
     //captionEnd = e;
   }
   public boolean filterGoal( GEDataModel g )
   {
     if( captionStart.length() > 0 )
     {
       if( captionStart.compareTo( g.getCaption() ) != 0 )
       {
         return true;
       }
     }
     /*if( captionEnd.length() > 0 )
     {
       if( captionEnd.compareTo( g.getCaption() ) != 0 )
       {
         return true;
       }
     }*/
     return false;
   }
}

class GEServiceComboBox extends JComboBox
{
   protected static ArrayList labels = new ArrayList();

   public static final GEServiceComboBoxFilter_None filterNo = new GEServiceComboBoxFilter_None();

   ///////////////////////////////////////////////////////////
   public GEServiceComboBox()
   {
   }
   public GEServiceComboBox( GEServiceComboBoxFilter filter )
   {
     setList( createList( filter ) );
   }

   ///////////////////////////////////////////////////////////
   public void setList( ArrayList list )
   {
     labels = list;

     super.removeAllItems();

     for( int i = 0; i < labels.size(); i++ )
     {
       addItem( labels.get( i ) );
     }
   }

   public ArrayList getList()
   {
     return labels;
   }

   ///////////////////////////////////////////////////////////
   public static ArrayList createList( GEServiceComboBoxFilter filter )
   {
     if( filter == null ) filter = new GEServiceComboBoxFilter_None();

     ArrayList captions = new ArrayList();

     //ArrayList goals = SortArrayList( GEApp.getDocument().getGoals() );
     ArrayList goals = SortArrayList( GEApp.getDocument().getDataList() );
     for( int i = 0; i < goals.size(); i++ )
     {
       GEDataModel g = (GEDataModel)goals.get( i );
       if( !filter.filterGoal( g ) ) captions.add( g.getCaption() );
     }

     return captions;
   }

   ///////////////////////////////////////////////////////////
   private static DuplicateMap getSortedMap( ArrayList list )
   {
     DuplicateMap map = new DuplicateMap();

     for( int i = 0; i < list.size(); i++ )
     {
       GEDataModel g = (GEDataModel)list.get( i );
       map.put( g.getCaption(), g );
     }
     return map;
   }

   ///////////////////////////////////////////////////////////
   public void selectGoal( GEService.DataModel s )
   {
     super.setSelectedItem( s.getCaption() );
   }

   ///////////////////////////////////////////////////////////
   private static ArrayList SortArrayList( ArrayList list )
   {
     DuplicateMap map = getSortedMap( list );

     ArrayList ret_list = new ArrayList();

     Set keys = map.keySet();
     for ( Iterator it = keys.iterator(); it.hasNext(); )
     {
       int n = 0;
       String key = ( String ) it.next();
       GEDataModel g = (GEDataModel)map.get( key );
       while( g != null )
       {
         ret_list.add( g );
         n++;
         g = (GEDataModel)map.get( key, n );
       }
     }
     return ret_list;
   }
}
