package goaleditor;

/*************************************************************************

  Copyright (c) 2005 Universita degli Studi di Trento All Rights Reserved.

This file contains software and/or modifications that are subject to the
University of Trento Shared Source License Version 1.0 (the 'License')
restricting usage for non-commercial purposes.

If this file contains source code, the licence should be available below
otherwise a copy of the License is available at http://www.dit.unitn.it/ssl

*************************************************************************/

/* Universita degli Studi di Trento ("UNITN") - Shared Source Licence Version 1.0 **

This License governs use of the accompanying Software, and your use of the Software
constitutes acceptance of this license.

You may use this Software for any non-commercial purpose, subject to the restrictions
in this license. Some purposes which can be non-commercial are teaching, academic
research, and personal experimentation. You may also distribute this Software with
books or other teaching materials, or publish the Software on websites, that are
intended to teach the use of the Software.

You may not use or distribute this Software or any derivative works in any form
for commercial purposes under this license. Examples of commercial purposes would be running business
operations, licensing, leasing, or selling the Software, or distributing the Software
for use with commercial products.

You may modify this Software and distribute the modified Software for non-commercial
purposes, however, you may not grant rights to the Software or derivative works that
are broader than those provided by this License. For example, you may not distribute
modifications of the Software under terms that would permit commercial use, or under
terms that purport to require the Software or derivative works to be sublicensed
to others.

You may use any information in intangible form that you remember after accessing
the Software. However, this right does not grant you a license to any of
UNITN's copyrights or patents for anything you might create using such information.

In return, we simply require that you agree:

- Not to remove any copyright or other notices from the Software.


- That if you distribute the Software in source or object form,
  you will include a verbatim copy of this license.


- That if you distribute derivative works of the Software in source code form
  you do so only under a license that includes all of the provisions of this
  License, and if you distribute derivative works of the Software solely in object
  form you do so only under a license that complies with this License.


- That if you have modified the Software or created derivative works, and distribute
  such modifications or derivative works, you will cause the modified files to carry
  prominent notices so that recipients know that they are not receiving the original
  Software. Such notices must state: (i) that you have changed the Software; and
  (ii) the date of any changes.


- THAT THE SOFTWARE COMES "AS IS", WITH NO WARRANTIES. THIS MEANS NO EXPRESS, IMPLIED
  OR STATUTORY WARRANTY, INCLUDING WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY
  OR FITNESS FOR A PARTICULAR PURPOSE OR ANY WARRANTY OF TITLE OR NON-INFRINGEMENT.
  ALSO, YOU MUST PASS THIS DISCLAIMER ON WHENEVER YOU DISTRIBUTE THE SOFTWARE OR
  DERIVATIVE WORKS.


- THAT UNITN WILL NOT BE LIABLE FOR ANY DAMAGES RELATED TO THE SOFTWARE OR THIS
  LICENSE, INCLUDING DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL OR INCIDENTAL DAMAGES,
  TO THE MAXIMUM EXTENT THE LAW PERMITS, NO MATTER WHAT LEGAL THEORY IT IS BASED ON.
  ALSO, YOU MUST PASS THIS LIMITATION OF LIABILITY ON WHENEVER YOU DISTRIBUTE THE
  SOFTWARE OR DERIVATIVE WORKS.


- That if you sue anyone over patents that you think may apply to the Software or
  anyone's use of the Software, your license to the Software ends automatically.


- That your rights under the License end automatically if you breach it in any way.


UNITN may, at its sole discretion, choose to license any modifications made by or
on behalf of UNITN under this License, or on different terms from those
contained in this license or may choose not to license them at all.

UNITN reserves all rights not expressly granted to you in this license.

********************************************************************************/

/**
 * @author Alberto Siena
 */

import java.awt.*;
import java.io.*;
import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
// DOM classes
import org.w3c.dom.*;
//JAXP 1.1
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.sax.*;


interface IXmlObject
{
  XmlNode createDescriptor();
  XmlNode getDescriptor();
  void setDescriptor( XmlNode nd );
}

class XmlNode implements CharSequence
{
  public String m_tag = "";
  public Map m_attr = new TreeMap();
  public DuplicateMap m_childs = new DuplicateMap();
  public String m_value = "";

  //static final XmlNode s_errorNode = new XmlNode( "", "#ERROR" );
  static final XmlNode s_noneNode = new XmlNode( "", "" );

  /*****************************************************/

  public static XmlNode create( String filename )
  {
    Document document = null;
    DocumentBuilder documentBuilder = null;
    DocumentBuilderFactory documentBuilderFactory = null;

    try
    {
      documentBuilderFactory = DocumentBuilderFactory.newInstance();
      documentBuilder = documentBuilderFactory.newDocumentBuilder();
      document = documentBuilder.parse( new File( filename ) );
    }
    catch(Exception e)
    {
      e.printStackTrace();

      return null;
    }

    return new XmlNode( document.getChildNodes().item( 0 ), true );
  }

  /*****************************************************/

  public XmlNode() {}
  public XmlNode( String tag )
  {
    m_tag = tag;
  }
  public XmlNode( String tag, String value )
  {
    m_tag = tag;
    m_value = value;
  }
  public XmlNode( String tag, String value, String[] attr_string )
  {
    m_tag = tag;
    m_value = value;
    for( int i = 0; i < attr_string.length; i += 2 )
    {
      setAttr( attr_string[i], attr_string[i+1] );
    }
  }
  public XmlNode( XmlNode copy )
  {
    m_tag = copy.getTag();
    m_value = copy.getValue();
    Set set = copy.enumAttributes();
    for( Iterator it = set.iterator(); it.hasNext(); )
    {
      String attr = (String)it.next();
      setAttr( attr, copy.getAttr( attr ) );
    }
  }
  public XmlNode( XmlNode copy, boolean copy_childs )
  {
    m_tag = copy.getTag();
    m_value = copy.getValue();
    Set set = copy.enumAttributes();
    for( Iterator it = set.iterator(); it.hasNext(); )
    {
      String attr = (String)it.next();
      setAttr( attr, copy.getAttr( attr ) );
    }
    if( !copy_childs ) return;

    XmlNodeList nodes = copy.getChilds();

    for( int i = 0; i < nodes.size(); i++ )
    {
      add( new XmlNode( nodes.node( i ), true ) );
    }
  }

  Document createDocument( File file )
  {
    Document document = null;
    DocumentBuilder documentBuilder = null;
    DocumentBuilderFactory documentBuilderFactory = null;

    try
    {
      documentBuilderFactory = DocumentBuilderFactory.newInstance();
      documentBuilder = documentBuilderFactory.newDocumentBuilder();
      document = documentBuilder.parse( file );
    }
    catch(Exception e)
    {
      e.printStackTrace();
      return null;
    }
    return document;
  }
  Document createDocument( InputStream is )
  {
    Document document = null;
    DocumentBuilder documentBuilder = null;
    DocumentBuilderFactory documentBuilderFactory = null;

    try
    {
      documentBuilderFactory = DocumentBuilderFactory.newInstance();
      documentBuilder = documentBuilderFactory.newDocumentBuilder();
      document = documentBuilder.parse( is );
    }
    catch( Exception ex )
    {
      ex.printStackTrace();
      return null;
    }
    return document;
  }

  public XmlNode( InputStream is )
  {
    Document document = createDocument( is );

    if( document == null ) return;

    add( document.getChildNodes().item( 0 ) );
  }
  public XmlNode( File file )
  {
    Document document = createDocument( file );

    if( document == null ) return;

    add( document.getChildNodes().item( 0 ) );
  }

  public XmlNode( Node nd )
  {
    add( nd );
  }
  public XmlNode( Node nd, boolean force )
  {
    m_tag = nd.getNodeName();

    NamedNodeMap attr = nd.getAttributes();

    for( int i = 0; i < attr.getLength(); i++ )
    {
      Node a = attr.item( i );
      setAttr( a.getNodeName(), a.getNodeValue() );
    }

    if( !hasSubElements( nd ) )
    {
      m_tag = m_tag.trim();
      m_value = getStringContent( nd );
      //add( m_tag.trim(), "Aaa" ); //getStringContent( nd ) );

      return;
    }

    NodeList nlist = nd.getChildNodes();

    for( int i = 0; i < nlist.getLength(); i++ )
    {
      Node item = nlist.item( i );

      if( item.getNodeType() == Node.TEXT_NODE ) continue;

      switch( item.getNodeType() )
      {
      case Node.ELEMENT_NODE:
        add( item.getNodeName(), new XmlNode( item, force ) );
        break;
      case Node.TEXT_NODE:
        //add( item.getNodeName().trim(), item.getFirstChild().getNodeValue() );
        break;
      default:
        break;
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////
  //  CharSequence

  public int length()
  {
    return m_value.length();
  }
  public char charAt(int index)
  {
    return m_value.charAt( index );
  }
  public CharSequence subSequence(int start, int end)
  {
    return m_value.subSequence( start, end );
  }
  public String toString()
  {
    return m_value.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////

  protected String printXmlNode( XmlNode node, int ind )
  {
    String txt = "";

    txt = indent( ind ) + "&lt;<font color=#aa0000>" + node.getTag() + "</font>";

    {
      Set attribs = node.enumAttributes();

      for ( Iterator it = attribs.iterator(); it.hasNext(); )
      {
        String key = (String)it.next();

        txt = txt + " <font color=#0000bb>" + key + "</font>=<font color=#008800>\"" + node.getAttr( key ) + "\"</font>";
      }
    }

    txt = txt + "&gt;";

    {
      ArrayList childs = node.getChilds();

      if( childs.size() > 0 )
      {
        txt = txt + "<br>";
        for ( int c = 0; c < childs.size(); c++ )
        {
          txt = txt + printXmlNode( ( XmlNode ) childs.get( c ), ind +1 );
        }
        txt = txt + indent( ind ) + "<font color=#999999>&lt;/" + node.getTag() + "&gt;</font><br>";
      }
      else
      {
        txt = txt + "<b>" + node.getValue().trim() + "</b>";
        txt = txt + "<font color=#999999>&lt;/" + node.getTag() + "&gt;</font><br>";
      }
    }

    return txt;
  }
  protected String indent( int ind )
  {
    String txt = "";
    for( int s = 0; s < ind; s++ ) txt = txt + "&nbsp;&nbsp;&nbsp;&nbsp;";
    return txt;
  }
  protected String indent( int ind, String t )
  {
    String txt = "";
    for( int s = 0; s < ind; s++ ) txt = txt + t;
    return txt;
  }
  public String getHtmlRepresentation()
  {
    return printXmlNode( this, 0 );
  }

  //////////////////////////////////////////////////////////////////////////////

  static class Painter
  {
    int m_textHeight = 10;
    int m_x = 0, m_y = 0;

    public void paint( XmlNode xml, Graphics g, int x, int y )
    {
      m_textHeight = g.getFontMetrics().getHeight();

      paintTag( xml, g, x, y + m_textHeight );
    }
    protected void paintTag( XmlNode tag, Graphics g, int x, int y )
    {
      m_x = x;
      m_y = y;

      g.setColor( Color.black );
      paintString( "<", g );

      g.setColor( Color.red );
      paintString( tag.getTag(), g );

      Set attr = tag.enumAttributes();
      for( Iterator it = attr.iterator(); it.hasNext(); )
      {
        String attrname = (String)it.next();

        g.setColor( Color.blue );
        paintString( " " + attrname, g );

        g.setColor( Color.black );
        paintString( "=\"" , g );

        g.setColor( Color.magenta );
        paintString( tag.getAttr( attrname ), g );

        g.setColor( Color.black );
        paintString( "\"", g );
      }
      g.setColor( Color.black );
      if( (tag.getValue().trim().length() == 0) && (tag.getChildCount() == 0) ) paintString( "/", g );
      paintString( ">" , g );

      {
        ArrayList childs = tag.getChilds();

        if( childs.size() > 0 )
        {
          m_y += m_textHeight;
          for( int i = 0; i < childs.size(); i++ )
          {
            XmlNode child = (XmlNode)childs.get( i );

            paintTag( child, g, x + 30, m_y );
          }
          m_x = x;
        }
        else
        {
          if( tag.getValue().trim().length() > 0 )
          {
            g.setColor( Color.black );
            paintString( tag.getValue(), g );
          }
          else
          {
            m_y += m_textHeight;
            return;
          }
        }
      }
      g.setColor( Color.gray );
      paintString( "</" + tag.getTag() + ">", g );

      m_y += m_textHeight;
    }
    protected void paintString( String s, Graphics g )
    {
      g.drawString( s, m_x, m_y );

      m_x += g.getFontMetrics().stringWidth( s );
    }

    public Dimension getPaintedSize()
    {
      return new Dimension( 100, m_y );
    }
  }

  public void paint( Graphics g, int x, int y )
  {
    new Painter().paint( this, g, x, y );
  }

  //////////////////////////////////////////////////////////////////////////////

  boolean hasSubElements( Node nd )
  {
    NodeList nlist = nd.getChildNodes();

    for( int i = 0; i < nlist.getLength(); i++ )
    {
      Node item = nlist.item( i );

      if( item.getNodeType() == Node.ELEMENT_NODE ) return true;
    }

    return false;
  }
  String getStringContent( Node nd )
  {
    NodeList nlist = nd.getChildNodes();
    String content = "";

    for( int i = 0; i < nlist.getLength(); i++ )
    {
      Node item = nlist.item( i );

      if( item.getNodeType() == Node.TEXT_NODE )
      {
        content = content + item.getNodeValue();
      }
    }

    return content.trim();
  }

  public void var_dump()
  {
    Set keys;

    System.out.println();

    System.out.println( "-----------" + this.hashCode() );

    System.out.println( ".VALORE: " + m_tag + " = " + getValue() );

    System.out.print( ".Attributi: " );

    keys = m_attr.keySet();
    for ( Iterator it = keys.iterator(); it.hasNext(); )
    {
      String key = ( String ) it.next();

      System.out.print( key + "=" + getAttr( key ) + " " );
    }

    System.out.println();

    keys = m_childs.keySet();
    for ( Iterator it = keys.iterator(); it.hasNext(); )
    {
      String key = ( String ) it.next();
      item( key ).var_dump();
    }
  }

  public boolean exists()
  {
    return (this != s_noneNode);
  }

  public void setValue( String value )
  {
    m_value = value;
  }
  public void setValue( boolean value )
  {
    m_value = new Boolean( value ).toString();
  }
  public void setValue( int n )
  {
    m_value = new Integer( n ).toString();
  }
  public void setValue( float f )
  {
    m_value = new Float( f ).toString();
  }
  public String getValue()
  {
    return m_value;
  }
  public String getValue( String def )
  {
    if( m_value.length() == 0 ) return def;
    return m_value;
  }
  ////////////////////////////////////////////
  public int getInt()
  {
    return Integer.parseInt( m_value );
  }
  public int getInt( int def )
  {
    int ret = def;

    try
    {
      ret = getInt();
    }
    catch( Exception ex ) {}

    return ret;
  }
  //////////////////////////////////////////////
  public float getFloat()
  {
    return Float.parseFloat( m_value );
  }
  public float getFloat( float def )
  {
    float ret = def;

    try
    {
      ret = getFloat();
    }
    catch( Exception ex ) {}

    return ret;
  }
  ////////////////////////////////////////////////
  public boolean getBoolean()
  {
    return Boolean.getBoolean( m_value.trim().toLowerCase() );
  }
  public boolean getBoolean( boolean def )
  {
    if( ! exists() ) return def;
    boolean ret = def;
    try
    {
      ret = (m_value.trim().toLowerCase().matches( "true" ) ? true : false);
    }
    catch( Exception ex ){}
    return ret;
  }
  ///////////////////////////////////////////////
  public String getTag()
  {
    return m_tag;
  }

  public void setAttr( String attr_name, String attr_value )
  {
    m_attr.put( attr_name, attr_value );
  }
  public String getAttr( String attr_name )
  {
    String ret = (String)m_attr.get( attr_name );
    if( ret == null ) return "";
    return ret;
  }
  public String getAttr( String attr_name, String def )
  {
    String ret = (String)m_attr.get( attr_name );
    if( ret == null ) return def;
    return ret;
  }
  public String formatAttrString()
  {
    String txt = "";
    Set attribs = enumAttributes();

    for ( Iterator it = attribs.iterator(); it.hasNext(); )
    {
      String key = (String)it.next();

      txt = txt + " " + key + "=\"" + getAttr( key ) + "\"";
    }

    return txt;
  }
  public Set enumAttributes()
  {
    return m_attr.keySet();
  }

  Set enumChildTags( String prepend )
  {
    Set childTags = new HashSet();

    for ( Iterator it = m_childs.keySet().iterator(); it.hasNext(); )
    {
      String key = ( String ) it.next();

      int n = this.getChildCount( key );

      if( n > 1 )
      {
        for( int i = 0; i < n; i++ )
        {
          childTags.add( prepend + key + ":" + i );
        }
      }
      else
      {
        childTags.add( prepend + key );
      }
    }

    return childTags;
  }

  public Set enumChildTags()
  {
    return enumChildTags( "" );
  }

  Set enumDescendantTags( String prepend )
  {
    Set tags = enumChildTags( prepend );
    XmlNodeList childs = getChilds();

    for( int i = 0; i < childs.size(); i++ )
    {
      tags.addAll( childs.node( i ).enumDescendantTags( prepend + childs.node( i ).getTag() + "/" ) );
    }

    return tags;
  }
  public Set enumDescendantTags()
  {
    return enumDescendantTags( "" );
  }

  public void add( Node nd )
  {
    m_tag = nd.getNodeName();

    NamedNodeMap attr = nd.getAttributes();

    for( int i = 0; i < attr.getLength(); i++ )
    {
      Node a = attr.item( i );
      setAttr( a.getNodeName(), a.getNodeValue() );
    }

    NodeList nlist = nd.getChildNodes();

    if( nlist.getLength() == 1 )
    {
      m_value = nlist.item( 0 ).getNodeValue();
      return;
    }

    for( int i = 0; i < nlist.getLength(); i++ )
    {
      Node item = nlist.item( i );

      if( item.getNodeType() == Node.ELEMENT_NODE )
      {
        if( item.getFirstChild() != null )
        {
          add( new XmlNode( item ) );
        }
      }
    }
  }
  public void add2( Node nd )
  {
    m_tag = nd.getNodeName();

    NamedNodeMap attr = nd.getAttributes();

    for( int i = 0; i < attr.getLength(); i++ )
    {
      Node a = attr.item( i );
      setAttr( a.getNodeName(), a.getNodeValue() );
    }

    //System.out.println( "XmlNode::add\r\n" + nd );

    NodeList nlist = nd.getChildNodes();

    for( int i = 0; i < nlist.getLength(); i++ )
    {
      Node item = nlist.item( i );

      if( item.getNodeType() == Node.ELEMENT_NODE )
      {
        if( item.getFirstChild() != null )
        {
          //if( item.getFirstChild().getNodeType() == Node.TEXT_NODE )
          {
            //System.out.println( item.getNodeName() + " = " + item.getFirstChild().getNodeValue() );
          }
          if( (item.getFirstChild().getNodeType() == Node.TEXT_NODE) ) //&& (nlist.getLength() == 1) )
          {
            add( item.getNodeName().trim(), item.getFirstChild().getNodeValue() );
          }
          else
          {
            add( item.getNodeName(), new XmlNode( item ) );
          }
        }
        else
        {
          add( item.getNodeName().trim(), "" );
        }
      }
    }
  }
  public void add( String key, XmlNode child )
  {
    m_childs.put( key, child );
  }
  public XmlNode add( XmlNode child )
  {
    m_childs.put( child.getTag(), child );

    return child;
  }
  public void add( String key, String value )
  {
    add( key, new XmlNode( key, value ) );
  }
  public void add( String key, double value )
  {
    add( key, new String( "" + value ) );
  }
  public void add( String key, float value )
  {
    add( key, new String( "" + value ) );
  }
  public void add( String key, int value )
  {
    add( key, new String( "" + value ) );
  }
  public void add( String key, boolean value )
  {
    add( key, new String( "" + value ) );
  }

  public XmlNode item( String key )
  {
    if( m_childs.get( key ) == null ) return XmlNode.s_noneNode;

    return (XmlNode)(m_childs.get( key ));
  }
  public XmlNode item( String key, boolean create )
  {
    if( m_childs.get( key ) == null )
    {
      if( create )
      {
        add( new XmlNode( key ) );
      }
      else
      {
        return XmlNode.s_noneNode;
      }
    }

    return (XmlNode)(m_childs.get( key ));
  }
  public XmlNode item( String key, int index )
  {
    if( m_childs.get( key, index ) == null ) return XmlNode.s_noneNode;

    return (XmlNode)(m_childs.get( key, index ));
  }
  public XmlNode item( String key, int index, boolean create )
  {
    if( create )
    {
      while( m_childs.get( key, index ) == null )
      {
        add( new XmlNode( key ) );
      }
    }

    if( m_childs.get( key, index ) == null )
    {
      return XmlNode.s_noneNode;
    }

    return (XmlNode)(m_childs.get( key, index ));
  }
  public XmlNode item( int index )
  {
    return (XmlNode)m_childs.get( index );
  }

  public XmlNode child( String path )
  {
    String tokens[] = path.split( "/" );

    if( tokens.length == 1 ) return item( path, true );

    XmlNode node = this;

    for( int i = 0; i < tokens.length; i++ )
    {
      String[] parts = tokens[i].split( ":" );

      if( parts.length == 1 )
      {
        node = node.item( tokens[i], true );
      }
      else if( parts.length == 2 )
      {
        node = node.item( parts[0], Integer.valueOf( parts[1] ).intValue(), true );
      }
    }

    return node;
  }

  public String get( String key )
  {
    return item( key ).getValue();
  }

  public XmlNode findByAttr( String attr, String value )
  {
    XmlNodeList childs = getChilds();
    int n = childs.size();
    for( int i = 0; i < n; i++ )
    {
      XmlNode node = childs.node( i );
      if( node.getAttr( attr ).compareTo( value ) == 0 ) return node;
    }
    return null;
  }

  public int childsCount()
  {
    return m_childs.size();
  }

  protected String getTextRepresentation( int ind )
  {
    String txt = "";

    txt = indent( ind, "    " ) + "<" + getTag() + "";

    {
      Set attribs = enumAttributes();

      for ( Iterator it = attribs.iterator(); it.hasNext(); )
      {
        String key = (String)it.next();

        txt = txt + " " + key + "=\"" + getAttr( key ) + "\"";
      }
    }

    txt = txt + ">";

    {
      ArrayList childs = getChilds();

      if( childs.size() > 0 )
      {
        txt = txt + "\r\n";
        for ( int c = 0; c < childs.size(); c++ )
        {
          txt = txt + ( (XmlNode) childs.get( c )).getTextRepresentation( ind +1 );
        }
        txt = txt + indent( ind, "    " ) + "</" + getTag() + ">\r\n";
      }
      else
      {
        txt = txt + getValue().trim();
        txt = txt + "</" + getTag() + ">\r\n";
      }
    }

    return txt;
  }

  public void write( File file )
  {
    try
    {
      PrintWriter out = new PrintWriter(new OutputStreamWriter( new FileOutputStream( file ),"UTF-8") );
      out.write( "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\" ?>\n" );
      out.write( getTextRepresentation( 0 ) );
      out.close();
    }
    catch( Exception ex ) {}
  }

  public void write( TransformerHandler hd )
  {
    try
    {
      //Set keys = m_data.keySet();
      AttributesImpl atts = new AttributesImpl();

      if( m_tag.trim().length() > 0 )
      {
        Set keys = m_attr.keySet();
        for ( Iterator it = keys.iterator(); it.hasNext(); )
        {
          String key = ( String ) it.next();

          atts.addAttribute( "", "", key, "CDATA", getAttr( key ) );
        }
        hd.startElement( "", "", m_tag, atts );
      }

      if( m_value.trim().length() > 0 )
      {
        hd.characters( m_value.toCharArray(), 0, m_value.length() );
      }

      Set keys = m_childs.keySet();
      for ( Iterator it = keys.iterator(); it.hasNext(); )
      {
        String key = ( String ) it.next();
        XmlNode content = item( key );
        int itn = 0;

        while( content.exists() )
        {
          atts.clear();
          //hd.startElement( "", "", key, atts );

          if ( content instanceof XmlNode )
          {
            ( ( XmlNode ) content ).write( hd );
          }
          else
          {
            String str = ( String ) m_childs.get( key );
            hd.characters( str.toCharArray(), 0, str.length() );
          }
          itn++;
          content = item( key, itn );
        }
        //hd.endElement( "", "", key );
      }
      if( m_tag.trim().length() > 0 )
      {
        hd.endElement( "", "", m_tag );
      }
    }
    catch ( Exception ex )
    {
      ex.printStackTrace();
    }
  }
  public boolean read( Node node )
  {
    return true;
  }

  public int getChildCount( String tagname )
  {
    return m_childs.count( tagname );
  }
  public int getChildCount()
  {
    return m_childs.size();
  }

  public XmlNodeList getChilds()
  {
    XmlNodeList ret = new XmlNodeList();

    ret.addAll( m_childs.valueSet() );

    /*for ( Iterator it = m_childs.keySet().iterator(); it.hasNext(); )
    {
      String key = ( String ) it.next();

      int n = this.getChildCount( key );

      for( int i = 0; i < n; i++ )
      {
        ret.add( m_childs.get( key, i ) );
      }
    }*/

    return ret;
  }

  public XmlNodeList getChilds( String tagname )
  {
    XmlNodeList ret = new XmlNodeList();

    int n = getChildCount( tagname );

    for( int i = 0; i < n; i++ )
    {
      ret.add( item( tagname, i ) );
    }

    return ret;
  }
}

class XmlNodeList extends ArrayList
{
  public XmlNode node( int n )
  {
    return (XmlNode)get( n );
  }
}